import itertools
from typing import Dict, Generator, List, NamedTuple, Set, Tuple
from collections import defaultdict
from heapq import heapify, heappop, heappush
import numpy as np

from numpy.linalg import norm
from sklearn.metrics import pairwise_distances
from sklearn.cluster import KMeans
from scipy.sparse.linalg import lsmr
from scipy.sparse import lil_array, hstack, csc_array
from enum import Enum
from queue import Queue
from .cell_complex import CellComplex, normalize_cell
from .spanning_trees import *

def project_flow(A: np.ndarray, flow: np.ndarray):
    """
    Projects the flow into Im(A).
    """
    flow_star, _, _, _, _, _, _, _ = lsmr(A, flow)
    return A @ flow_star

def score_cell_flows(cell_compl: CellComplex, flows: List[np.array], cell_boundary: csc_array) -> np.ndarray:
    """
    Scores the individual flows, i.e., calculates the element-wise sum of squares of the approximation error for each flow.
    """
    old_boundary = cell_compl.boundary_map(2)
    new_boundary = hstack((old_boundary, cell_boundary))

    scores = np.zeros((len(flows),), dtype=np.float64)
    for idx, flow in enumerate(flows):
        harmonic = flow - project_flow(new_boundary, flow)
        scores[idx] = np.sum(np.square(harmonic))
    return scores

def score_cells_multiple(cell_compl: CellComplex, flows: List[np.array], cells: List[Tuple[tuple, csc_array]], length_penalty: float = 0.01) -> tuple[np.ndarray, list[tuple]]:
    """
    Scores multiple cells on all given flows.
    """
    no_boundary = csc_array((len(cell_compl.get_cells(1)), 0))
    empty_cell = ((), no_boundary)
    cell_idx = [cell[0] for cell in cells + [empty_cell]]
    results = [score_cell_flows(cell_compl, flows, cell[1]) + length_penalty * len(cell)
               for cell in cells + [empty_cell]]
    return np.array(results), cell_idx

def inv_edge(e: Tuple[int, int]):
    return (e[1], e[0])

def get_induced_cycle(edge: Tuple[int, int], st: FlowPotentialSpanningTree) -> tuple:
    """
    Gets the cycle induced by adding edge to the spanning tree modeled by node_level and parent_node
    """
    left = []
    right = []

    a = edge[0]
    b = edge[1]

    if st.node_level[a] < st.node_level[b]:
        a = edge[1]
        b = edge[0]

    while st.node_level[a] > st.node_level[b]:
        left.append(a)
        a = st.parent_node[a]

    while a != b:
        left.append(a)
        right.append(b)
        a = st.parent_node[a]
        b = st.parent_node[b]

    left.append(a)
    return normalize_cell(tuple(left + right[::-1]))

class CellCandidateHeuristic(Enum):
    """
    Methods to find potential cells in cell complexes.
    DFS: depth-first-search to construct a spanning tree, choosing the cell 
         with most circle flow from all cycles induced by adding a single edge to the spanning tree.
    BFS: breadth-first-search (see DFS for details)
    MAX: maximum spanning tree (by `sum(abs(edge_flows))`)
    SIMILARITY: cluster by harmonic flow values, then build spanning tree with edges that have maximum similarity to each cluster center
    TRIANGLES: consider all triangles
    GROUND_TRUTH: consider all true cells
    """
    DFS = 1
    BFS = 2
    MAX = 3
    SIMILARITY = 4
    TRIANGLES = 5


class CellSearchFlowNormalization(Enum):
    """
    How to normalize the integrated cell flows (score = integrated flow / normalization)

    NONE: Don't normalize (i.e., take complete integrated flow)
    LEN: Length of the cell. Efficiently implemented, does not increase asymptotic complexity.
    """
    NONE = 1
    LEN = 2


def find_max_cycles(cell_compl: CellComplex, spanning_tree: FlowPotentialSpanningTree, flows: np.ndarray,
                    edges: List[Tuple[int, int, int]],
                    flow_norm: CellSearchFlowNormalization = CellSearchFlowNormalization.LEN,
                    cycle_flows_heap: list | None = None) -> List[Tuple[float, FlowPotentialSpanningTree, int, int, int]]:
    """
    Checks all cycles induced by `edges` on `spanning_tree` and calculates the total flow (w.r.t `flow_norm`)

    returns a heap of tuples (-normalized_flow, edge_idx, node_a, node_b)  
    Note that the implementation of `heapq` is a min-heap, therefore the negative of the flow is added.  
    However, typically, the flow is only required to `heappop`.
    """
    if cycle_flows_heap is None:
        cycle_flows_heap = []
    # empty or already a heap, so not necessary: heapify(cycle_flows_heap)

    node_potential = spanning_tree.node_potential
    node_level = spanning_tree.node_level
    edge_count = len(cell_compl.get_cells(1))

    # edge -> least common ancestor
    lca: Dict[Tuple[int, int], int]

    if flow_norm == CellSearchFlowNormalization.LEN:
        # O(nodes + edges) operation, so only perform if necessary
        lca = spanning_tree.least_common_ancestor([edge[1:] for edge in edges])

    def add_cycle(edge: int, node: int, other: int, edge_flows: np.ndarray):
        total_flow = np.sum(
            np.abs(node_potential[node] - edge_flows - node_potential[other]))
        if flow_norm == CellSearchFlowNormalization.LEN:
            total_flow /= node_level[node] + node_level[other] + 1 - 2 * node_level[lca[(node, other)]]
        heappush(cycle_flows_heap, (-total_flow, spanning_tree, edge, node, other))

    for edge, node, other in edges:
        edge_flows = flows[:, edge % edge_count] * (-1) ** (edge // edge_count)
        add_cycle(edge, node, other, edge_flows)

    return cycle_flows_heap

def __finalize_candidates(cell_compl: CellComplex, heap: list[tuple[float, FlowPotentialSpanningTree, int, int, int]], 
                        count: int) -> list[tuple[tuple, csc_array, FlowPotentialSpanningTree]]:
    """
    Takes a heap resulting from a candidate search on one or more `FlowPotentialSpanningTree`s and returns a list of candidates for evaluation.

    Also de-duplicates.
    """
    result = []
    found_cycles = set()

    while len(result) < count:
        _, spanning_tree, _, u, v = heappop(heap)
        cycle = get_induced_cycle((u,v), spanning_tree)
        if cycle not in found_cycles:
            result.append((cycle, cell_compl.calc_cell_boundary(cycle), spanning_tree))
            found_cycles.add(cycle)
    
    return result

def cell_candidate_search_st(rnd: np.random.Generator, cell_compl: CellComplex, result_count: int, flows: np.ndarray, method: CellCandidateHeuristic = CellCandidateHeuristic.SIMILARITY,
                            n_clusters = 11, flow_norm: CellSearchFlowNormalization = CellSearchFlowNormalization.NONE) -> List[Tuple[tuple, csc_array, FlowPotentialSpanningTree]]:
    """
    Searches for Cell candidates using spanning trees.
    The concrete implementation depends on `method`:
    BFS: randomly select `result_count` breadth-first-search spanning trees and get the cell with most integrated flow from each
    DFS: randomly select `result_count` depth-first-search spanning trees and get the cell with most integrated flow from each
    MAX: construct the maximum spanning tree (according to `sum(abs(flows))`) and get the `result_count` cells with the most integrated flow.
    see `cell_candidate_search_rnd_st` and `cell_candidate_search_max_st` respectively for more info.
    """
    max_cycle_heap = []
    if method == CellCandidateHeuristic.MAX:
        spanning_tree, edges = max_spanning_tree(cell_compl, flows)

        max_cycle_heap = find_max_cycles(cell_compl, spanning_tree, flows, edges, flow_norm)
    elif method == CellCandidateHeuristic.SIMILARITY:
        # since the orientation of edges is arbitrary, we need to cluster both with and opposite the orientation
        embed = np.concatenate((flows, -flows), axis=1).T
        kmeans = KMeans(n_clusters=n_clusters, n_init='auto', random_state=rnd.integers(4294967295)).fit(embed)
        dist_center_edges_pos = pairwise_distances(kmeans.cluster_centers_, flows.T)
        dist_center_edges_neg = pairwise_distances(kmeans.cluster_centers_, -flows.T)
        dist_center_edges = np.minimum(dist_center_edges_pos, dist_center_edges_neg)
        for idx, similarity in enumerate(dist_center_edges):
            spanning_tree, edges = max_spanning_tree(cell_compl, flows, -1 * similarity, st_id = idx)

            find_max_cycles(cell_compl, spanning_tree, flows, edges, flow_norm, max_cycle_heap)
    else:
        incidences = cell_compl.node_incidences()
        nodes = np.fromiter(incidences.keys(), dtype=int)
        edge_count = len(cell_compl.get_cells(1))

        for _ in range(result_count):
            root_node = rnd.choice(nodes)
            if method == CellCandidateHeuristic.BFS:
                spanning_tree, edges = bfs_spanning_tree(rnd, incidences, edge_count, root_node, flows)
            else: # CellCandidateHeuristic.DFS
                spanning_tree, edges = dfs_spanning_tree(rnd, incidences, edge_count, root_node, flows)

            find_max_cycles(cell_compl, spanning_tree, flows, edges, flow_norm, max_cycle_heap)
    return __finalize_candidates(cell_compl, max_cycle_heap, result_count)

def cell_candidate_search_triangles(cell_compl: CellComplex, result_count: int, flows: np.ndarray) -> List[Tuple[tuple, csc_array, FlowPotentialSpanningTree]]:
    """
    Returns the most significant triangles according to the flow around their edges.
    """
    cand_list = [(np.sum(np.abs(boundary.T @ flows.T)), cell, boundary) for cell, boundary in cell_compl.triangles]
    cand_list.sort(key=lambda x: x[0], reverse=True)
    return [(cell, boundary) for _, cell, boundary in cand_list[:result_count]]

def cell_inference_approximation(CC: CellComplex, flows: np.ndarray, n_candidates: int,
                                 n: int | None = None, epsilon: float | None = None, heuristic: CellCandidateHeuristic = CellCandidateHeuristic.SIMILARITY,
                                 n_clusters: int = 11, flow_norm = CellSearchFlowNormalization.LEN,
                                 seed: np.random.Generator | int | None = None) -> CellComplex:
    """
    Runs the inference and approximation algorithm from [1] and returns both the inferred flows and the resulting CellComplex.

    Parameters:

    - `CC`: The CellComplex to start from
    - `flows`: numpy array of shape (flow samples, edges). The edges dimension needs to be ordered corresponding to `CC.get_cells(1)`.
    - `n_candidates`: The number of cell candidates to consider in each step.
    - `n`: Number of cells to add. Optional, see for termination condition below.
    - `epsilon`: Approximation error to reach before stopping. Optional, see for termination condition below.
    - `heuristic`: The cell inference heuristic. Optional, default: `CellCandidateHeuristic.SIMILARITY`
    - `n_clusters`: Number of clusters for `heuristic=CellCandidateHeuristic.SIMILARITY`. Otherwise ignored. Intuitively, it can be t to 1 + 2 * (Number of cells to distinguish in each step) Optional, default: `11`.
    - `flow_norm`: How to normalize cell flows for comparison. See CellSearchFlowNormalization. Optional, default `CellSearchFlowNormalization.LEN`.
    - `seed`: Seed for reproducible Randomness. Can be a `np.random.Generator`, `int` (used to seed a Generator). Optional, `np.random.default_rng()` if not provided.

    Returns: The cell complex obtained by adding the mentioned cells.

    Terminates on either of these conditions:

    - the number of added cells is `n`
    - the approximation error is at most `epsilon`
    - the best cell candidate does not result in a lower approximation error than previously (with warning)

    Recommended usage:

    - Only provide `n` to add a certain number of cells
    - If you want to reach a certain `epsilon`, you may want to provide a very large `n` to ensure it still terminates.
    - It is possible, but not recommended, to provide neither. In this case, the algorithm will only terminate once it fails to add another cell that is linearly independent from previous cells. Not considering flaws in the heuristic or numeric computation, this results in adding a complete cycle basis.

    [1]: Josef Hoppe and Michael T. Schaub: Representing Edge Flows on Graphs via Sparse Cell Complexes. In: arXiv pre-prints, 2023. https://arxiv.org/abs/2309.01632
    """
    if seed is None:
        rnd = np.random.default_rng()
    elif type(seed) is np.random.Generator:
        rnd = seed
    else:
        rnd = np.random.default_rng(seed)

    if n is None and epsilon is None:
        print("[WARN] running cell detection without explicit termination condition.")
    
    current_compl = CC
    no_gradient_flows = np.copy(flows)
    for i in range(flows.shape[0]):
        no_gradient_flows[i] -= project_flow(
            current_compl.boundary_map(1).T, flows[[i]])
    
    harmonic_flows = np.copy(no_gradient_flows)
    for j in range(harmonic_flows.shape[0]):
        harmonic_flows[j] -= project_flow(
            current_compl.boundary_map(2), no_gradient_flows[[j]])

    cells_added = 0
    approx_error = np.sqrt(np.sum(np.square(harmonic_flows)))

    while (n is None or cells_added < n) and (epsilon is None or approx_error > epsilon):
        if heuristic == CellCandidateHeuristic.TRIANGLES:
            candidate_cells = cell_candidate_search_triangles(current_compl, n_candidates, harmonic_flows)
        else:
            candidate_cells = cell_candidate_search_st(rnd, current_compl, n_candidates, harmonic_flows,
                                                       heuristic, n_clusters, flow_norm)

        score_vals, score_cells = score_cells_multiple(
            current_compl, no_gradient_flows, [cell[:2] for cell in candidate_cells])
        scores = np.sum(score_vals, axis=1)
        next_cell = score_cells[np.argmin(scores)]
        
        cell_boundaries = { cell[0]: cell[1] for cell in candidate_cells}

        if next_cell == ():
            print(
                f'[WARN] detected empty cell after {cells_added} cells (error: {approx_error:.2G})')
            return current_compl

        current_compl = current_compl.add_cell_fast(next_cell, cell_boundaries[next_cell])

        harmonic_flows = np.copy(no_gradient_flows)
        for j in range(harmonic_flows.shape[0]):
            harmonic_flows[j] -= project_flow(
                current_compl.boundary_map(2), no_gradient_flows[[j]])
        
        approx_error = np.sum(np.square(harmonic_flows))
        cells_added += 1

    return current_compl


"""
Cell FLOWer. Implements algorithms and requirements from [1].

[1]: Josef Hoppe and Michael T. Schaub: Representing Edge Flows on Graphs via Sparse Cell Complexes. In: arXiv pre-prints, 2023. https://arxiv.org/abs/2309.01632
"""

from .cell_complex import CellComplex, normalize_cell, nx_graph_to_cc, cc_to_nx_graph, map_cells_to_index, cell_to_index, index_to_cell
from .detection import CellCandidateHeuristic, CellSearchFlowNormalization, cell_inference_approximation


"""
Some helpers for linalg / matrix operations
"""

import numpy as np
import numpy.linalg as linalg


def normalize(matrix: np.ndarray) -> np.ndarray:
    interm = np.diag(np.diag(np.ones(matrix.shape).T @ matrix))
    return matrix @ linalg.inv(interm)

def neg_part(matrix: np.ndarray) -> np.ndarray:
    return -(np.minimum(matrix, np.zeros(matrix.shape)))

def pos_part(matrix: np.ndarray) -> np.ndarray:
    return np.maximum(matrix, np.zeros(matrix.shape))



import cell_flower as cf
import pandas as pd
import numpy as np
import networkx as nx

def test_basic_run():
    # Initialize underlying graph
    CC = cf.CellComplex(5, [(0,1), (1,2), (2,3), (0,3), (3,4), (0,4)])

    # Cell boundary maps expressed as pd.Series
    # Cell (0,1,2,3)
    square_cell = pd.Series({
            (0,1): 1,
            (1,2): 1,
            (2,3): 1,
            (0,3): -1
        }, index=CC.get_cells(1)).fillna(0).sort_index()

    # Cell (0,3,4)
    triangle_cell = pd.Series({
            (0,3): 1,
            (3,4): 1,
            (0,4): -1
        }, index=CC.get_cells(1)).fillna(0).sort_index()

    # Flows

    flow1 = square_cell * 1 - triangle_cell * .5
    flow2 = square_cell * .3 + triangle_cell * .6

    flows = np.array([flow1.to_numpy(), flow2.to_numpy()])

    CC_prime = cf.cell_inference_approximation(CC, flows, 2, 2, n_clusters=5)
    # Check to see the cells are correct
    assert (0,1,2,3) in CC_prime.get_cells(2)
    assert (0,3,4) in CC_prime.get_cells(2)

def test_nx_compatibility():
    G = nx.karate_club_graph()
    CC, _, _ = cf.nx_graph_to_cc(G)
    G2 = cf.cc_to_nx_graph(CC)
    assert nx.is_isomorphic(G, G2)


"""
Functions to find and handle spanning trees.
"""

from collections import defaultdict
from typing import NamedTuple, Tuple, Dict, Set, List
import numpy as np
from queue import Queue

from .cell_complex import CellComplex

class FlowPotentialSpanningTree(NamedTuple):
    """
    A spanning tree with (multi-dimensional) potentials on nodes
    """
    st_id: int
    parent_node: np.ndarray
    node_level: np.ndarray
    node_potential: np.ndarray

    def least_common_ancestor(self, node_pairs: List[Tuple[int,int]]) -> Dict[Tuple[int,int], int]:
        """
        Implementation of Tarjan's off-line lowest common ancestors algorithm
        (https://en.wikipedia.org/wiki/Tarjan%27s_off-line_lowest_common_ancestors_algorithm, https://doi.org/10.1145%2F322154.322161)

        Returns edge (as tuple (a,b)) -> ancestor

        The time complexity is in O(n + m * α(n + m)) for n = |nodes|, m = |node_pairs|
        (α is the inverse of the Ackermann function f(x) = A(x,x))

        Note that Tarjan later gave an improved algorithm with complexity O(n+m).
        Since the Ackermann function grows so fast, α is almost constant.
        It is especially faster than the sorting we perform afterward.
        """
        node_color = np.zeros_like(self.parent_node, dtype=np.int8)
        ancestor = np.arange(self.parent_node.shape[0])
        partition = UnionFind(node_color.shape[0])

        # need to traverse the spanning tree top-down
        children = defaultdict(list)
        root = -1
        for node, parent in enumerate(self.parent_node):
            if node != parent:
                children[parent].append(node)
            else:
                root = node

        # need to efficiently retrieve queries
        # node -> (other, edge[0], edge[1])
        queries = defaultdict(list)
        for (a,b) in node_pairs:
            queries[a].append((b,a,b))
            queries[b].append((a,a,b))

        result = {}
        
        def lca(node):
            for child in children[node]:
                lca(child)
                partition.unite(node, child)
                ancestor[partition[node]] = node
            node_color[node] = 1
            for other, a,b in queries[node]:
                if node_color[other] == 1:
                    result[(a,b)] = ancestor[partition[other]]
        
        lca(root)

        return result

class UnionFind:
    """
    Starts with `items` different sets and supports two operations:
    1. get unique identifier for set (via __getitem__). The identifier is also part of the set.
    2. unite two sets

    Both operations work in O(α(n)) (with amortisation - individual operations may take longer)
    (α is the inverse of the Ackermann function, so practically constant.)
    """
    parent: np.ndarray

    def __init__(self, items: int) -> None:
        self.parent = np.arange(items)

    def __getitem__(self, idx: int) -> int:
        parent = self.parent[idx]
        # check if parent assigned correctly
        if self.parent[parent] == parent:
            return parent
        # find correct root anestor
        while self.parent[parent] != parent:
            parent = self.parent[parent]
        # update all ancestors to point to root
        p = idx
        while p != parent:
            next_p = self.parent[p]
            self.parent[p] = parent
            p = next_p
        return parent

    def unite(self, a: int, b: int):
        """
        unites the set containing a with the set containing b
        """
        self.parent[self[a]] = self[b]


def max_spanning_tree(cell_compl: CellComplex, flows: np.ndarray,
                      weight: np.ndarray | None = None, st_id = 0) ->\
        Tuple[FlowPotentialSpanningTree, List[Tuple[int,int,int]]]:
    """
    Finds the maximal spanning tree according to `sum(abs(flows))`

    weight: edges are added from highest do lowest weight. default `np.sum(np.abs(flows)`

    Returns: (SpanningTree, List[(edge_idx, node, node)])
    """
    tree_incidences: Dict[int, Set[Tuple[int, int]]] = defaultdict(set)
    non_tree_edges = []

    if weight is None:
        weight = np.sum(np.abs(flows), axis=0)

    edge_list = cell_compl.get_cells(1)
    max_weight_edges = np.argsort(weight)[::-1]
    conn_comps = UnionFind(len(cell_compl.get_cells(0)))

    for edge_idx in max_weight_edges:
        a, b = edge_list[edge_idx]
        if conn_comps[a] != conn_comps[b]:
            # add current edge
            tree_incidences[a].add((b, edge_idx))
            tree_incidences[b].add((a, edge_idx + len(edge_list)))
            conn_comps.unite(a, b)
        else:
            non_tree_edges.append((edge_idx, a, b))

    tree, additional_edges = bfs_spanning_tree(
        rnd=None, node_incidences=tree_incidences, edge_count=len(edge_list), root_node=0, flows=flows, st_id = st_id)
    
    if len(additional_edges) > 0:
        # should never happen, just to catch errors
        raise RuntimeError('Max ST is not a tree')


    return tree, non_tree_edges


def dfs_spanning_tree(rnd: np.random.Generator | None,
                      node_incidences: Dict[int, Set[Tuple[int, int]]], edge_count: int, root_node: int,
                      flows: np.ndarray) -> Tuple[FlowPotentialSpanningTree, List[Tuple[int,int,int]]]:
    """
    Finds a spanning tree via depth-first-search.
    if `rnd` is supplied, it is used to randomize the order in which children are traversed.

    `node_incidences` has the same format as AbstractComplex.node_incidences:
    node -> (neighbor, edge_idx)
    where len(edges) is added to edge_idx if the traversal is from the higher node index to the lower.

    returns (spanning_tree, List[(edge_idx, node, node)])
    """
    permutation = rnd.permutation if rnd is not None else lambda x: x

    nodes = len(node_incidences.keys())
    # spanning tree properties
    parent_node = np.full((nodes,), -1, dtype=np.int32)
    parent_node[root_node] = root_node
    node_level = np.full((nodes,), -1, dtype=np.int32)
    node_level[root_node] = 0
    node_potential = np.zeros((nodes, flows.shape[0]), dtype=np.float64)

    non_tree_edges = []

    def dfs(node):
        for (other, edge) in permutation(np.fromiter(node_incidences[node], dtype=tuple)):
            if other != parent_node[node]:
                edge_flows = flows[:, edge % edge_count] * \
                    (-1) ** (edge // edge_count)
                if parent_node[other] == -1:
                    # undiscovered
                    parent_node[other] = node
                    node_level[other] = node_level[node] + 1
                    node_potential[other] = node_potential[node] - edge_flows
                    dfs(other)
                else:
                    non_tree_edges.append((edge % edge_count, node, other))
    dfs(root_node)

    return FlowPotentialSpanningTree(parent_node, node_level, node_potential), non_tree_edges


def bfs_spanning_tree(rnd: np.random.Generator | None,
                      node_incidences: Dict[int, Set[Tuple[int, int]]], edge_count: int, root_node: int,
                      flows: np.ndarray, st_id = 0) -> Tuple[FlowPotentialSpanningTree, List[Tuple[int,int,int]]]:
    """
    Finds a spanning tree via depth-first-search.
    if `rnd` is supplied, it is used to randomize the order in which children are traversed.

    `node_incidences` has the same format as AbstractComplex.node_incidences:
    node -> (neighbor, edge_idx)
    where len(edges) is added to edge_idx if the traversal is from the higher node index to the lower.

    returns (spanning_tree, List[(edge_idx, node, node)])
    """
    permutation = rnd.permutation if rnd is not None else lambda x: x

    nodes = max(node_incidences.keys()) + 1
    # spanning tree properties
    parent_node = np.full((nodes,), -1, dtype=np.int32)
    parent_node[root_node] = root_node
    node_level = np.full((nodes,), -1, dtype=np.int32)
    node_level[root_node] = 0
    node_potential = np.zeros((nodes, flows.shape[0]), dtype=np.float64)

    non_tree_edges = []

    def bfs(root):
        queue = Queue()
        queue.put(root)
        while not queue.empty():
            node = queue.get()
            # randomize because the BFS tree with root n is not unique:
            #  root
            #  /  \
            # a -- b
            #  \  / <-- two possible BFS trees: in one, c is the child of a, in the other it is the child of b
            #   c
            for (other, edge) in permutation(np.fromiter(node_incidences[node], dtype=tuple)):
                if other != parent_node[node]:
                    edge_flows = flows[:, edge %
                                       edge_count] * (-1) ** (edge // edge_count)
                    if parent_node[other] == -1:
                        # undiscovered
                        parent_node[other] = node
                        node_level[other] = node_level[node] + 1
                        node_potential[other] = node_potential[node] - \
                            edge_flows
                        queue.put(other)
                    else:
                        non_tree_edges.append((edge % edge_count, node, other))
    bfs(root_node)

    return FlowPotentialSpanningTree(st_id, parent_node, node_level, node_potential), non_tree_edges


from itertools import combinations
from typing import DefaultDict, Any, Literal, Callable

import numpy as np
from scipy.sparse import csc_array, lil_array, hstack

def normalize_cell(cell: tuple) -> tuple:
    """
    Transforms the cell (represented as a sequence of nodes) to normalized form.

    Normalized form: `cell_prime := (m, a, [...], b)` s.t.:
    
    - `m := min(cell)`
    - `a < b`
    - `cell_prime` represents the same cycle as `cell`

    returns `cell_prime`
    """
    min_index = cell.index(min(cell))
    shifted = cell[min_index:] + cell[:min_index]
    if shifted[-1] < shifted[1]:
        shifted = shifted[::-1]
        shifted = shifted[-1:] + shifted[:-1]
    return shifted

def calc_edges(cell: tuple) -> list[tuple]:
    """
    Calculates all edges as 2-tuples of nodes from a 2-dim cell
    """
    res = []
    for i, node in enumerate(cell):
        j = (i + 1) % len(cell)
        res.append(normalize_cell((node, cell[j])))
    return res

def cell_to_index(cell: tuple, node_index_map: dict[Any, int]) -> tuple:
    """Maps the given cell tuple to an int tuple according to the node index map.

    Usage: see `map_cells_to_index`.
    
    Returns: tuple of ints representing the cell.
    """
    return normalize_cell(tuple([node_index_map[n] for n in cell]))

def index_to_cell(cell: tuple, node_list: list[Any]) -> tuple:
    """Maps the given index tuple to a cell tuple according to the node list.

    Usage: see `map_cells_to_index`.
    
    Returns: tuple of original objects representing the cell.
    """
    return tuple([node_list[i] for i in cell])

def map_cells_to_index(cells: list[tuple], sort_key: Literal[None] | Callable[[Any], Any] = None) -> tuple[int, list[tuple], list[Any], dict[Any,int]]:
    """Maps the given cell tuples to int tuples and a dictionary for reverse mapping

    Usage:

    ```python
    import cell_flower as cf
    cells = [('A', 'B'), ('B','C','D')]
    node_count, cells, node_list, node_index_map = cf.map_cells_to_int(cells)
    CC = cf.CellComplex(node_count, cells)
    print(cf.index_to_cell((0,1,2), node_list)) # ('A', 'B', 'C') [ordered like `normalize_cell` for usage in CC]
    print(cf.cell_to_index(('A', 'B', 'C'), node_list)) # (0, 1, 2) [ordered as given for external usage]
    ```

    Parameters:
    - `cells`: List of all cells (as tuples of any type). May include 0-cells as tuples of length 1.
    - `sort_key`: Sort key used to sort nodes by passing it as the `key` parameter to `sorted()`. If `None`, comparable types will still be sorted.
    
    Returns: number of nodes, list of 1-cells and 2-cells as tuples, ordered list of nodes, dictionary to map nodes to int representation
    """
    node_list = list(set().union(*cells))
    if sort_key != None:
        node_list = sorted(node_list, key=sort_key)
    else:
        try:
            node_list = sorted(node_list)
        except:
            pass
    node_index_map = {
        node: idx for node, idx in enumerate(node_list)
    }
    index_cells = [ cell_to_index(cell, node_index_map) for cell in cells if len(cell) > 1 ]
    return len(node_list), index_cells, node_list, node_index_map

class CellComplex():
    """
    Implementation of a cell complex

    Only supports cells of dimension 0, 1, or 2 (i.e., nodes, edges, polygons)
    """
    cell_order_map: dict[int, list[tuple[int]]]
    __cell_index: dict[tuple[int], int]
    embedding = None
    __cell_boundary_map: csc_array
    __edge_boundary_map: csc_array
    __node_incidences: dict[int, dict[int, int]]
    __triangles: np.ndarray[tuple[tuple,csc_array]] | None = None

    @property
    def triangles(self) -> np.ndarray[tuple[tuple,csc_array]]:
        """
        All triangles in the complex.
        """
        if self.__triangles is None:
            self.__calc_triangles()
        return self.__triangles
    
    def __calc_triangles(self):
        candidate_set = set()

        edge_set = set(self.get_cells(1))
        incidences = self.node_incidences()

        for node, neighbors in incidences.items():
            for (a, _), (b, _) in combinations(neighbors, 2):
                if normalize_cell((a,b)) in edge_set:
                    candidate_set.add(normalize_cell((node,a,b)))

        self.__triangles = np.ndarray(len(candidate_set), dtype=tuple)
        
        for idx, cycle in enumerate(candidate_set):
            self.__triangles[idx] = (cycle, self.calc_cell_boundary(cycle))

    def skeleton(self, order: int = 1) -> "CellComplex":
        """
        Returns the skeleton of the complex.

        Note: Currently, only the 1-skeleton is implemented efficiently.
        """
        if order == 1:
            cell_order_map = DefaultDict(list)
            cell_index = DefaultDict(lambda: {})
            for i in range(order + 1):
                cell_order_map[i] += self.cell_order_map[i]
                cell_index[i] = self.__cell_index[i].copy()

            empty_boundaries = csc_array((len(self.get_cells(1)), 0), dtype=np.int32)
            return CellComplex(None, cell_order_map=cell_order_map, 
                    cell_boundary_map=empty_boundaries, node_incidences=self.__node_incidences,
                    edge_boundary_map=self.__edge_boundary_map, cell_index=cell_index)
        # very slow
        # todo implement fast version for other orders
        cells = []
        for i in range(order + 1):
            cells += self.cell_order_map[i]
        return CellComplex(cells)

    def add_cell(self, cell: tuple) -> "CellComplex":
        """
        Returns a cell complex with `cell` added.

        Note: the implementation is currently not very efficient, use `add_cell_fast` to add 2-cells efficiently.
        """
        cells = []
        for i in range(1,3):
            cells += self.cell_order_map[i]
        cells += [cell]
        return CellComplex(len(self.cell_order_map[0]), cells)
    
    def add_cell_fast(self, cell:tuple, cell_boundary: csc_array) -> "CellComplex":
        """
        Adds a cell efficiently if the boundary is given.

        Returns a new cell complex with the added cell.
        """
        if len(cell) < 3:
            raise NotImplementedError('Can only add 2-cells')
        cell_order_map = DefaultDict(list)
        cell_index = DefaultDict(lambda: {})
        for i in range(3):
            cell_order_map[i] += self.cell_order_map[i]
            cell_index[i] = self.__cell_index[i].copy()
        cell_order_map[2].append(cell)
        cell_index[2][cell] = len(cell_order_map[2]) - 1
        new_compl = CellComplex(-1, [], cell_order_map=cell_order_map, 
                cell_boundary_map=hstack((self.__cell_boundary_map, cell_boundary)),
                edge_boundary_map=self.__edge_boundary_map, node_incidences=self.__node_incidences, cell_index=cell_index)
        if self.__triangles is not None:
            new_compl.__triangles = self.__triangles
        return new_compl

    def __init__(self, node_count: int, cells: list[tuple[int]], **kwargs):
        """
        Initializes a new Cell complex, either by calculating everything (if node_count != -1) or by taking all properties from **kwargs
        """
        if node_count == -1:
            # 'manual' initialization (see also: add_cell_fast)
            self.cell_order_map = kwargs['cell_order_map']
            self.__cell_boundary_map = kwargs['cell_boundary_map']
            self.__edge_boundary_map = kwargs['edge_boundary_map']
            self.__node_incidences = kwargs['node_incidences']
            self.__cell_index = kwargs['cell_index']
            return
        self.cell_order_map = DefaultDict(list)
        self.__cell_index = DefaultDict(lambda: {})

        def add_cell(order, cell):
            if not cell in self.__cell_index[order]:
                self.cell_order_map[order].append(cell)
                self.__cell_index[order][cell] = len(self.cell_order_map[order]) - 1
        
        def assert_cell(order, cell):
            if not cell in self.__cell_index[order]:
                raise RuntimeError(f"Unknown but required cell {cell}. Please use `cf.map_cells_to_index()`.")

        for node in range(node_count):
            # TODO: Refactor to replace 0-cells with node_count attribute?
            add_cell(0, (node,))

        for cell in cells:
            if len(cell) == 1:
                assert_cell(0, cell)
            elif len(cell) == 2:
                assert_cell(0, (cell[0],))
                assert_cell(0, (cell[1],))
                add_cell(1, cell)
            else:
                norm_cell = normalize_cell(cell)
                add_cell(2, norm_cell)
                for point in norm_cell:
                    assert_cell(0, (point,))
                for edge in calc_edges(norm_cell):
                    add_cell(1, edge)
        
        # calculate cell boundaries
        edge_count = len(self.get_cells(1))
        cell_count = len(self.get_cells(2))
        cell_boundary = lil_array((edge_count, cell_count), dtype=np.int32)
        for upper_idx, cell in enumerate(self.get_cells(2)):
            for i, _ in enumerate(cell):
                j = (i + 1) % len(cell)
                lower_idx = self.__cell_index[1][tuple(sorted([cell[i], cell[j]]))]
                orientation = 1 if cell[i] < cell[j] else -1
                cell_boundary[lower_idx, upper_idx] = orientation
        self.__cell_boundary_map = csc_array(cell_boundary)

        edge_boundary = lil_array((node_count, edge_count), dtype=np.int32)
        for upper_idx, edge in enumerate(self.get_cells(1)):
            edge_boundary[self.__cell_index[0][(edge[0],)], upper_idx] = -1
            edge_boundary[self.__cell_index[0][(edge[1],)], upper_idx] = 1
        self.__edge_boundary_map = csc_array(edge_boundary)
        self.__node_incidences = self.__node_incidences()

    def calc_cell_boundary(self, cell: tuple) -> csc_array:
        """
        Get the boundary map for a hypothetical cell. Re-calculates the boundary map, so use the existing boundary map for existing cells.
        """
        edge_count = len(self.get_cells(1))
        cell_boundary = lil_array((edge_count, 1), dtype=np.int32)
        for i, _ in enumerate(cell):
            j = (i + 1) % len(cell)
            lower_idx = self.__cell_index[1][tuple(sorted([cell[i], cell[j]]))]
            orientation = 1 if cell[i] < cell[j] else -1
            cell_boundary[lower_idx, 0] = orientation
        return csc_array(cell_boundary)

    def get_cells(self, k: int = 1) -> list:
        """
        returns the cells $C_k$ of the given dimension `k`.
        """
        return self.cell_order_map[k]

    def boundary_map(self, k: int = 1) -> np.ndarray | csc_array:
        """The boundary map $B_k$. Sparse for $k \in {1,2}$."""
        if k == 1:
            return self.__edge_boundary_map
        if k == 2:
            return self.__cell_boundary_map
        if k == 0:
            return np.zeros(shape=(0,len(self.get_cells(k))), dtype=np.int32)
        raise NotImplementedError('Only cells up to dimension 2 (nodes, edges, polygons) are supported.')
    
    def node_incidences(self) -> dict[int, set[tuple[int, int]]]:
        """
        dictionary `node` -> `edge`; edges are represented as tuples of nodes
        """
        return self.__node_incidences
    
    def __node_incidences(self) -> dict[int, set[tuple[int, int]]]:
        """
        For all nodes, get all adjacent nodes and the index of the connecting edge.
        ```
                            ↓ index of the edge. len(edges) is added to the index iff opposite to edge orientation
        per node: (other, edge_index)
                    ↑ other node (as number / name)
        ```
        """
        res = DefaultDict(set)
        edge_count = len(self.get_cells(1))
        for idx, (a, b) in enumerate(self.get_cells(1)):
            res[a].add((b, idx))
            res[b].add((a, idx + edge_count))
        return res
    
def cc_to_nx_graph(CC: CellComplex):
    """
    Converts the 1-skeleton of the given CC to the networkx.Graph format.

    Requires networkx to be on the path.
    """
    import networkx as nx
    G = nx.Graph()
    # add separately to ensure correct order
    G.add_nodes_from([c[0] for c in CC.get_cells(0)])
    G.add_edges_from(CC.get_cells(1))
    return G

def nx_graph_to_cc(G) -> tuple[CellComplex, list[Any], dict[Any,int]]:
    """
    Converts the networkx Graph to a Cell Complex.

    Always index-converts the given nodes using `map_cells_to_index`, consult that function for more info.
    Also returns the node list and node index map from `map_cells_to_index`.

    Returns: Tuple (cell complex, ordered list of nodes, dictionary to map nodes to int representation)
    """
    cells = [(node, ) for node in G.nodes] + [e for e in G.edges]
    node_count, idx_cells, node_list, node_index_map = map_cells_to_index(cells)
    return CellComplex(node_count, idx_cells), node_list, node_index_map